!MNH_LIC Copyright 1994-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
MODULE MODI_ICE4_FAST_RS
INTERFACE
SUBROUTINE ICE4_FAST_RS(LDSOFT, PCOMPUTE, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, &
                       &PLBDAR, PLBDAS, &
                       &PT,  PRVT, PRCT, PRRT, PRST, &
                       &PRIAGGS, &
                       &PRCRIMSS, PRCRIMSG, PRSRIMCG, &
                       &PRRACCSS, PRRACCSG, PRSACCRG, PRSMLTG, &
                       &PRCMLTSR, &
                       &PRS_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RS, PA_RG)
IMPLICIT NONE
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIAGGS  ! r_i aggregation on r_s
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCRIMSS ! Cloud droplet riming of the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCRIMSG ! Cloud droplet riming of the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSRIMCG ! Cloud droplet riming of the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRACCSS ! Rain accretion onto the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRACCSG ! Rain accretion onto the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSACCRG ! Rain accretion onto the aggregates
REAL, DIMENSION(:),       INTENT(INOUT) :: PRSMLTG  ! Conversion-Melting of the aggregates
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCMLTSR ! Cloud droplet collection onto aggregates by positive temperature
REAL, DIMENSION(:, : ),   INTENT(INOUT) :: PRS_TEND ! Individual tendencies
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
END SUBROUTINE ICE4_FAST_RS
END INTERFACE
END MODULE MODI_ICE4_FAST_RS


SUBROUTINE ICE4_FAST_RS(LDSOFT, PCOMPUTE, &
                       &PRHODREF, PLVFACT, PLSFACT, PPRES, &
                       &PDV, PKA, PCJ, &
                       &PLBDAR, PLBDAS, &
                       &PT,  PRVT, PRCT, PRRT, PRST, &
                       &PRIAGGS, &
                       &PRCRIMSS, PRCRIMSG, PRSRIMCG, &
                       &PRRACCSS, PRRACCSG, PRSACCRG, PRSMLTG, &
                       &PRCMLTSR, &
                       &PRS_TEND, &
                       &PA_TH, PA_RC, PA_RR, PA_RS, PA_RG)
!!
!!**  PURPOSE
!!    -------
!!      Computes the fast rs processes
!!
!!    AUTHOR
!!    ------
!!      S. Riette from the splitting of rain_ice source code (nov. 2014)
!!
!!    MODIFICATIONS
!!    -------------
!!
!  P. Wautelet 26/04/2019: replace non-standard FLOAT function by REAL function
!  P. Wautelet 29/05/2019: remove PACK/UNPACK intrinsics (to get more performance and better OpenACC support)
!
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST,            ONLY: XALPI,XALPW,XBETAI,XBETAW,XCI,XCL,XCPV,XESTT,XGAMI,XGAMW,XLMTT,XLVTT,XMD,XMV,XRV,XTT,  &
                               XEPSILO
USE MODD_PARAM_ICE,      ONLY: LEVLIMIT, CSNOWRIMING
USE MODD_RAIN_ICE_DESCR, ONLY: XBS,XCEXVT,XCXS,XRTMIN
USE MODD_RAIN_ICE_PARAM, ONLY: NACCLBDAR,NACCLBDAS,NGAMINC,X0DEPS,X1DEPS,XACCINTP1R,XACCINTP1S,XACCINTP2R,XACCINTP2S, &
                               XCRIMSG,XCRIMSS,XEX0DEPS,XEX1DEPS,XEXCRIMSG,XEXCRIMSS,XEXSRIMCG,XEXSRIMCG2,XFRACCSS,   &
                               XFSACCRG,XFSCVMG,XGAMINC_RIM1,XGAMINC_RIM2,XGAMINC_RIM4,XKER_RACCS,                    &
                               XKER_RACCSS,XKER_SACCRG,XLBRACCS1,XLBRACCS2,XLBRACCS3,XLBSACCR1,XLBSACCR2,XLBSACCR3,   &
                               XRIMINTP1,XRIMINTP2,XSRIMCG,XSRIMCG2,XSRIMCG3

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK,      ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
USE MODE_MPPDB

#ifdef MNH_BITREP
USE MODI_BITREP
#endif
!
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
LOGICAL,                  INTENT(IN)    :: LDSOFT
REAL, DIMENSION(:),       INTENT(IN)    :: PCOMPUTE
REAL, DIMENSION(:),       INTENT(IN)    :: PRHODREF ! Reference density
REAL, DIMENSION(:),       INTENT(IN)    :: PLVFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PLSFACT
REAL, DIMENSION(:),       INTENT(IN)    :: PPRES    ! absolute pressure at t
REAL, DIMENSION(:),       INTENT(IN)    :: PDV      ! Diffusivity of water vapor in the air
REAL, DIMENSION(:),       INTENT(IN)    :: PKA      ! Thermal conductivity of the air
REAL, DIMENSION(:),       INTENT(IN)    :: PCJ      ! Function to compute the ventilation coefficient
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAR   ! Slope parameter of the raindrop  distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PLBDAS   ! Slope parameter of the aggregate distribution
REAL, DIMENSION(:),       INTENT(IN)    :: PT       ! Temperature
REAL, DIMENSION(:),       INTENT(IN)    :: PRVT     ! Water vapor m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRCT     ! Cloud water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRRT     ! Rain water m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRST     ! Snow/aggregate m.r. at t
REAL, DIMENSION(:),       INTENT(IN)    :: PRIAGGS  ! r_i aggregation on r_s
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCRIMSS ! Cloud droplet riming of the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRCRIMSG ! Cloud droplet riming of the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSRIMCG ! Cloud droplet riming of the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRACCSS ! Rain accretion onto the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRRACCSG ! Rain accretion onto the aggregates
REAL, DIMENSION(:),       INTENT(OUT)   :: PRSACCRG ! Rain accretion onto the aggregates
REAL, DIMENSION(:),       INTENT(INOUT) :: PRSMLTG  ! Conversion-Melting of the aggregates
REAL, DIMENSION(:),       INTENT(INOUT) :: PRCMLTSR ! Cloud droplet collection onto aggregates by positive temperature
REAL, DIMENSION(:, : ),   INTENT(INOUT) :: PRS_TEND ! Individual tendencies
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_TH
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RC
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RR
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RS
REAL, DIMENSION(:),       INTENT(INOUT) :: PA_RG
!
!*       0.2  declaration of local variables
!
INTEGER, PARAMETER :: IRCRIMS=1, IRCRIMSS=2, IRSRIMCG=3, IRRACCS=4, IRRACCSS=5, IRSACCRG=6, &
                    & IFREEZ1=7, IFREEZ2=8
!
INTEGER                            :: IGRIM, IGACC
INTEGER                            :: ISIZE
INTEGER                            :: IDX, JJ, JL
INTEGER, DIMENSION(:), POINTER, CONTIGUOUS :: I1
INTEGER, DIMENSION(:), POINTER, CONTIGUOUS :: IVEC1, IVEC2
LOGICAL, DIMENSION(:), POINTER, CONTIGUOUS :: GRIM, GACC
REAL,    DIMENSION(:), POINTER, CONTIGUOUS :: ZRIM, ZACC, ZMASK
REAL,    DIMENSION(:), POINTER, CONTIGUOUS :: ZVEC1, ZVEC2, ZVEC3
REAL,    DIMENSION(:), POINTER, CONTIGUOUS :: ZZW, ZZW2, ZZW6, ZFREEZ_RATE

!$acc data present( PCOMPUTE, PRHODREF, PLVFACT, PLSFACT, PPRES, PDV, PKA, PCJ,                        &
!$acc&              PLBDAR, PLBDAS, PT, PRVT, PRCT, PRRT, PRST, PRIAGGS, PRCRIMSS, PRCRIMSG, PRSRIMCG, &
!$acc&              PRRACCSS, PRRACCSG, PRSACCRG, PRSMLTG, PRCMLTSR, PRS_TEND, PA_TH,                  &
!$acc&              PA_RC, PA_RR, PA_RS, PA_RG )                                                       &
!$acc&     present( XGAMINC_RIM1, XGAMINC_RIM2, XGAMINC_RIM4 )
!
!-------------------------------------------------------------------------------
!
IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK1D(PCOMPUTE,"ICE4_FAST_RS beg:PCOMPUTE")
  CALL MPPDB_CHECK1D(PRHODREF,"ICE4_FAST_RS beg:PRHODREF",PRECISION)
  CALL MPPDB_CHECK1D(PLVFACT,"ICE4_FAST_RS beg:PLVFACT",PRECISION)
  CALL MPPDB_CHECK1D(PLSFACT,"ICE4_FAST_RS beg:PLSFACT",PRECISION)
  CALL MPPDB_CHECK1D(PPRES,"ICE4_FAST_RS beg:PPRES",PRECISION)
  CALL MPPDB_CHECK1D(PDV,"ICE4_FAST_RS beg:PDV",PRECISION)
  CALL MPPDB_CHECK1D(PKA,"ICE4_FAST_RS beg:PKA",PRECISION)
  CALL MPPDB_CHECK1D(PCJ,"ICE4_FAST_RS beg:PCJ",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAR,"ICE4_FAST_RS beg:PLBDAR",PRECISION)
  CALL MPPDB_CHECK1D(PLBDAS,"ICE4_FAST_RS beg:PLBDAS",PRECISION)
  CALL MPPDB_CHECK1D(PT,"ICE4_FAST_RS beg:PT",PRECISION)
  CALL MPPDB_CHECK1D(PRVT,"ICE4_FAST_RS beg:PRVT",PRECISION)
  CALL MPPDB_CHECK1D(PRCT,"ICE4_FAST_RS beg:PRCT",PRECISION)
  CALL MPPDB_CHECK1D(PRRT,"ICE4_FAST_RS beg:PRRT",PRECISION)
  CALL MPPDB_CHECK1D(PRST,"ICE4_FAST_RS beg:PRST",PRECISION)
  CALL MPPDB_CHECK1D(PRIAGGS,"ICE4_FAST_RS beg:PRIAGGS",PRECISION)
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRSMLTG,"ICE4_FAST_RS beg:PRSMLTG",PRECISION)
  CALL MPPDB_CHECK1D(PRCMLTSR,"ICE4_FAST_RS beg:PRCMLTSR",PRECISION)
  CALL MPPDB_CHECK2D(PRS_TEND,"ICE4_FAST_RS beg:PRS_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RS beg:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RS beg:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RS beg:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RS beg:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RS beg:PA_RG",PRECISION)
END IF


ISIZE = Size( PRHODREF )

#ifndef MNH_OPENACC
allocate( i1          ( isize ) )
allocate( ivec1       ( isize ) )
allocate( ivec2       ( isize ) )
allocate( grim        ( isize ) )
allocate( gacc        ( isize ) )
allocate( zrim        ( isize ) )
allocate( zacc        ( isize ) )
allocate( zmask       ( isize ) )
allocate( zvec1       ( isize ) )
allocate( zvec2       ( isize ) )
allocate( zvec3       ( isize ) )
allocate( zzw         ( isize ) )
allocate( zzw2        ( isize ) )
allocate( zzw6        ( isize ) )
allocate( zfreez_rate ( isize ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( I1,    ISIZE )
CALL MNH_MEM_GET( IVEC1, ISIZE )
CALL MNH_MEM_GET( IVEC2, ISIZE )

CALL MNH_MEM_GET( grim, ISIZE )
CALL MNH_MEM_GET( gacc, ISIZE )

CALL MNH_MEM_GET( zrim,        ISIZE )
CALL MNH_MEM_GET( zacc,        ISIZE )
CALL MNH_MEM_GET( zmask,       ISIZE )
CALL MNH_MEM_GET( zvec1,       ISIZE )
CALL MNH_MEM_GET( zvec2,       ISIZE )
CALL MNH_MEM_GET( zvec3,       ISIZE )
CALL MNH_MEM_GET( zzw,         ISIZE )
CALL MNH_MEM_GET( zzw2,        ISIZE )
CALL MNH_MEM_GET( zzw6,        ISIZE )
CALL MNH_MEM_GET( zfreez_rate, ISIZE )
#endif
!
!$acc data present( I1, IVEC1, IVEC2, GRIM, GACC, ZRIM, ZACC, ZMASK, ZVEC1, ZVEC2, ZVEC3, ZZW, ZZW2, ZZW6, ZFREEZ_RATE ) &
!$acc&     copyin ( XRTMIN )

!
!*       5.0    maximum freezing rate
!
!$acc kernels copyout(igrim)
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(5)-PRST(JL))) * & ! WHERE(PRST(:)>XRTMIN(5))
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRS_TEND(JL, IFREEZ1)=ZMASK(JL) * PRS_TEND(JL, IFREEZ1)
    PRS_TEND(JL, IFREEZ2)=ZMASK(JL) * PRS_TEND(JL, IFREEZ2)
  ENDDO
ELSE
  DO JL=1, ISIZE
    PRS_TEND(JL, IFREEZ1)=ZMASK(JL) * PRVT(JL)*PPRES(JL)/(XEPSILO+PRVT(JL)) ! Vapor pressure
  ENDDO
  IF(LEVLIMIT) THEN
    WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
      PRS_TEND(:, IFREEZ1)=MIN(PRS_TEND(:, IFREEZ1), EXP(XALPI-XBETAI/PT(:)-XGAMI*ALOG(PT(:)))) ! min(ev, es_i(T))
#else
      PRS_TEND(:, IFREEZ1)=MIN(PRS_TEND(:, IFREEZ1), BR_EXP(XALPI-XBETAI/PT(:)-XGAMI*BR_LOG(PT(:)))) ! min(ev, es_i(T))
#endif
    END WHERE
  ENDIF
  PRS_TEND(:, IFREEZ2)=0.
  WHERE(ZMASK(:)==1.)
    PRS_TEND(:, IFREEZ1)=PKA(:)*(XTT-PT(:)) +                              &
             (PDV(:)*(XLVTT+(XCPV-XCL)*(PT(:)-XTT)) &
                           *(XESTT-PRS_TEND(:, IFREEZ1))/(XRV*PT(:))           )
#ifndef MNH_BITREP
    PRS_TEND(:, IFREEZ1)=PRS_TEND(:, IFREEZ1)* ( X0DEPS*       PLBDAS(:)**XEX0DEPS +     &
                           X1DEPS*PCJ(:)*PLBDAS(:)**XEX1DEPS )/ &
#else
    PRS_TEND(:, IFREEZ1)=PRS_TEND(:, IFREEZ1)* ( X0DEPS*       BR_POW(PLBDAS(:),XEX0DEPS) +     &
                           X1DEPS*PCJ(:)*BR_POW(PLBDAS(:),XEX1DEPS) )/ &
#endif
                          ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )
    PRS_TEND(:, IFREEZ2)=(PRHODREF(:)*(XLMTT+(XCI-XCL)*(XTT-PT(:)))   ) / &
                          ( PRHODREF(:)*(XLMTT-XCL*(XTT-PT(:))) )
  END WHERE
ENDIF
!$acc loop independent
DO JL=1, ISIZE
  !We must agregate, at least, the cold species
  !And we are only interested by the freezing rate of liquid species
  ZFREEZ_RATE(JL)=ZMASK(JL) * MAX(0., MAX(0., PRS_TEND(JL, IFREEZ1) + &
                                              &PRS_TEND(JL, IFREEZ2) * PRIAGGS(JL)) - &
                                      PRIAGGS(JL))
ENDDO
!$acc end kernels
!*       5.1    cloud droplet riming of the aggregates
!
IGRIM = 0
!$acc parallel loop private(idx) copy(IGRIM) independent present_cr( I1, GRIM, PCOMPUTE, PRCT, PRST, XRTMIN, ZRIM )
DO JJ = 1 , SIZE( GRIM ) 
  ZRIM(JJ)=MAX(0., -SIGN(1., XRTMIN(2)-PRCT(JJ))) * & !WHERE(PRCT(:)>XRTMIN(2))
          &MAX(0., -SIGN(1., XRTMIN(5)-PRST(JJ))) * & !WHERE(PRST(:)>XRTMIN(5))
          &PCOMPUTE(JJ)
  IF (ZRIM(JJ)>0) THEN
!$acc atomic capture
    IGRIM = IGRIM + 1
    IDX = IGRIM
!$acc end atomic
    I1(IDX) = JJ
    GRIM(JJ) = .TRUE.
  ELSE
    GRIM(JJ) = .FALSE.
  END IF
END DO
! Collection of cloud droplets by snow: this rate is used for riming (T<0) and for conversion/melting (T>0)
IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRS_TEND(JL, IRCRIMS)=ZRIM(JL) * PRS_TEND(JL, IRCRIMS)
    PRS_TEND(JL, IRCRIMSS)=ZRIM(JL) * PRS_TEND(JL, IRCRIMSS)
    PRS_TEND(JL, IRSRIMCG)=ZRIM(JL) * PRS_TEND(JL, IRSRIMCG)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRS_TEND(:, IRCRIMS)=0.
  PRS_TEND(:, IRCRIMSS)=0.
  PRS_TEND(:, IRSRIMCG)=0.
!$acc end kernels
  !
  IF(IGRIM>0) THEN
    !
    !        5.1.1  select the PLBDAS
    !
!$acc kernels copyin( CSNOWRIMING )
    !$mnh_do_concurrent( JJ = 1 : IGRIM )
      ZVEC1(JJ) = PLBDAS(I1(JJ))
    !
    !        5.1.2  find the next lower indice for the PLBDAS in the geometrical
    !               set of Lbda_s used to tabulate some moments of the incomplete
    !               gamma function
    !
      ZVEC2(JJ) = MAX( 1.00001, MIN( REAL(NGAMINC)-0.00001,           &
#ifndef MNH_BITREP
                          XRIMINTP1 * LOG( ZVEC1(JJ) ) + XRIMINTP2 ) )
#else
                          XRIMINTP1 * BR_LOG( ZVEC1(JJ) ) + XRIMINTP2 ) )
#endif
      IVEC2(JJ) = INT( ZVEC2(JJ) )
      ZVEC2(JJ) = ZVEC2(JJ) - REAL( IVEC2(JJ) )
      !
      !        5.1.3  perform the linear interpolation of the normalized
      !               "2+XDS"-moment of the incomplete gamma function
      !
      ZVEC1(JJ) =   XGAMINC_RIM1( IVEC2(JJ)+1 )* ZVEC2(JJ)      &
                  - XGAMINC_RIM1( IVEC2(JJ)   )*(ZVEC2(JJ) - 1.0)
    !$mnh_end_do()
!$acc end kernels
!$acc kernels
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGRIM
      ZZW(I1(JJ)) = ZVEC1(JJ)
    END DO
    !
    !        5.1.4  riming of the small sized aggregates
    !
    !$mnh_do_concurrent( JJ = 1 : ISIZE )
      IF ( GRIM(JJ) ) THEN
        PRS_TEND(JJ, IRCRIMSS) = XCRIMSS * ZZW(JJ) * PRCT(JJ)                & ! RCRIMSS
#ifndef MNH_BITREP
                                         *   PLBDAS(JJ)**XEXCRIMSS &
                                         * PRHODREF(JJ)**(-XCEXVT)
#else
                                         * BR_POW(PLBDAS(JJ),XEXCRIMSS) &
                                         * BR_POW(PRHODREF(JJ),-XCEXVT)
#endif
      END IF
    !$mnh_end_do()
    !
    !        5.1.5  perform the linear interpolation of the normalized
    !               "XBS"-moment of the incomplete gamma function (XGAMINC_RIM2) and
    !               "XBG"-moment of the incomplete gamma function (XGAMINC_RIM4)
    !
    ZVEC1(1:IGRIM) =  XGAMINC_RIM2( IVEC2(1:IGRIM)+1 )* ZVEC2(1:IGRIM)      &
                    - XGAMINC_RIM2( IVEC2(1:IGRIM)   )*(ZVEC2(1:IGRIM) - 1.0)
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGRIM
      ZZW(I1(JJ)) = ZVEC1(JJ)
    END DO

    ZVEC1(1:IGRIM) =  XGAMINC_RIM4( IVEC2(1:IGRIM)+1 )* ZVEC2(1:IGRIM)      &
                    - XGAMINC_RIM4( IVEC2(1:IGRIM)   )*(ZVEC2(1:IGRIM) - 1.0)
    ZZW2(:) = 0.
!$acc loop independent
    DO JJ = 1, IGRIM
      ZZW2(I1(JJ)) = ZVEC1(JJ)
    END DO
    !
    !        5.1.6  riming-conversion of the large sized aggregates into graupeln
    !
    !
    !$mnh_do_concurrent( JJ = 1 : ISIZE )
      IF ( GRIM(JJ) ) THEN
        PRS_TEND(JJ, IRCRIMS)=XCRIMSG * PRCT(JJ)               & ! RCRIMS
#ifndef MNH_BITREP
                                     * PLBDAS(JJ)**XEXCRIMSG  &
                                     * PRHODREF(JJ)**(-XCEXVT)
#else
                                     * BR_POW(PLBDAS(JJ),XEXCRIMSG) &
                                     * BR_POW(PRHODREF(JJ),-XCEXVT)
#endif
        ZZW6(JJ) = PRS_TEND(JJ, IRCRIMS) - PRS_TEND(JJ, IRCRIMSS) ! RCRIMSG
      END IF
    !$mnh_end_do()

    IF(CSNOWRIMING=='M90 ')THEN
      !Murakami 1990
      WHERE(GRIM(:))
#ifndef MNH_BITREP
        PRS_TEND(:, IRSRIMCG)=XSRIMCG * PLBDAS(:)**XEXSRIMCG*(1.0-ZZW(:))
        PRS_TEND(:, IRSRIMCG)=ZZW6(:)*PRS_TEND(:, IRSRIMCG)/ &
                       MAX(1.E-20, &
                           XSRIMCG3*XSRIMCG2*PLBDAS(:)**XEXSRIMCG2*(1.-ZZW2(:)) - &
                           XSRIMCG3*PRS_TEND(:, IRSRIMCG))
#else
        PRS_TEND(:, IRSRIMCG)=XSRIMCG * BR_POW(PLBDAS(:),XEXSRIMCG)*(1.0-ZZW(:))
        PRS_TEND(:, IRSRIMCG)=ZZW6(:)*PRS_TEND(:, IRSRIMCG)/ &
                       MAX(1.E-20, &
                           XSRIMCG3*XSRIMCG2*BR_POW(PLBDAS(:),XEXSRIMCG2)*(1.-ZZW2(:)) - &
                           XSRIMCG3*PRS_TEND(:, IRSRIMCG))
#endif
      END WHERE
    ELSE
      PRS_TEND(:, IRSRIMCG)=0.
    END IF
!$acc end kernels
  ENDIF
ENDIF
!
!$acc kernels copyout(igacc)
!$acc loop independent
DO JL=1, ISIZE
  ! More restrictive RIM mask to be used for riming by negative temperature only
  ZRIM(JL)=ZRIM(JL) * &
          &MAX(0., -SIGN(1., PT(JL)-XTT)) ! WHERE(PT(:)<XTT)
  PRCRIMSS(JL)=ZRIM(JL)*MIN(ZFREEZ_RATE(JL), PRS_TEND(JL, IRCRIMSS))
  ZFREEZ_RATE(JL)=MAX(0., ZFREEZ_RATE(JL)-PRCRIMSS(JL))
  ZZW(JL) = MIN(1., ZFREEZ_RATE(JL) / MAX(1.E-20, PRS_TEND(JL, IRCRIMS) - PRCRIMSS(JL))) ! proportion we are able to freeze
  PRCRIMSG(JL) = ZRIM(JL) * ZZW(JL) * MAX(0., PRS_TEND(JL, IRCRIMS) - PRCRIMSS(JL)) ! RCRIMSG
  ZFREEZ_RATE(JL)=MAX(0., ZFREEZ_RATE(JL)-PRCRIMSG(JL))
  PRSRIMCG(JL) = ZRIM(JL) * ZZW(JL) * PRS_TEND(JL, IRSRIMCG)

  PRSRIMCG(JL) = PRSRIMCG(JL) * MAX(0., -SIGN(1., -PRCRIMSG(JL)))
  PRCRIMSG(JL)=MAX(0., PRCRIMSG(JL))

  PA_RC(JL) = PA_RC(JL) - PRCRIMSS(JL)
  PA_RS(JL) = PA_RS(JL) + PRCRIMSS(JL)
  PA_TH(JL) = PA_TH(JL) + PRCRIMSS(JL)*(PLSFACT(JL)-PLVFACT(JL))
  PA_RC(JL) = PA_RC(JL) - PRCRIMSG(JL)
  PA_RS(JL) = PA_RS(JL) - PRSRIMCG(JL)
  PA_RG(JL) = PA_RG(JL) + PRCRIMSG(JL)+PRSRIMCG(JL)
  PA_TH(JL) = PA_TH(JL) + PRCRIMSG(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!$acc end kernels
!
!*       5.2    rain accretion onto the aggregates
!
IGACC = 0
!$acc parallel loop private(idx) copy(IGACC) independent present_cr( I1, GACC, PCOMPUTE, PRRT, PRST, XRTMIN, ZACC )
DO JJ = 1 , SIZE( GACC ) 
  ZACC(JJ)=MAX(0., -SIGN(1., XRTMIN(3)-PRRT(JJ))) * & !WHERE(PRRT(:)>XRTMIN(3))
          &MAX(0., -SIGN(1., XRTMIN(5)-PRST(JJ))) * & !WHERE(PRST(:)>XRTMIN(5))
          &PCOMPUTE(JJ)
  IF (ZACC(JJ)>0) THEN
!$acc atomic capture
    IGACC = IGACC + 1
    IDX = IGACC
!$acc end atomic
    I1(IDX) = JJ
    GACC(JJ) = .TRUE.
  ELSE
    GACC(JJ) = .FALSE.
  END IF
END DO
IF(LDSOFT) THEN
!$acc kernels
  DO JL=1, ISIZE
    PRS_TEND(JL, IRRACCS)=ZACC(JL) * PRS_TEND(JL, IRRACCS)
    PRS_TEND(JL, IRRACCSS)=ZACC(JL) * PRS_TEND(JL, IRRACCSS)
    PRS_TEND(JL, IRSACCRG)=ZACC(JL) * PRS_TEND(JL, IRSACCRG)
  ENDDO
!$acc end kernels
ELSE
!$acc kernels
  PRS_TEND(:, IRRACCS)=0.
  PRS_TEND(:, IRRACCSS)=0.
  PRS_TEND(:, IRSACCRG)=0.
!$acc end kernels
  IF(IGACC>0)THEN
!$acc kernels
    !
    !
    !        5.2.1  select the (PLBDAS,PLBDAR) couplet
    !
!$acc loop independent
    DO JJ = 1, IGACC
      ZVEC1(JJ) = PLBDAS(I1(JJ))
      ZVEC2(JJ) = PLBDAR(I1(JJ))
    END DO
    !
    !        5.2.2  find the next lower indice for the PLBDAS and for the PLBDAR
    !               in the geometrical set of (Lbda_s,Lbda_r) couplet use to
    !               tabulate the RACCSS-kernel
    !
    !$mnh_do_concurrent( JJ = 1 : IGACC )
      ZVEC1(JJ) = MAX( 1.00001, MIN( REAL(NACCLBDAS)-0.00001,           &
#ifndef MNH_BITREP
                          XACCINTP1S * LOG( ZVEC1(JJ) ) + XACCINTP2S ) )
#else
                          XACCINTP1S * BR_LOG( ZVEC1(JJ) ) + XACCINTP2S ) )
#endif
      IVEC1(JJ) = INT( ZVEC1(JJ) )
      ZVEC1(JJ) = ZVEC1(JJ) - REAL( IVEC1(JJ) )
      !
      ZVEC2(JJ) = MAX( 1.00001, MIN( REAL(NACCLBDAR)-0.00001,           &
#ifndef MNH_BITREP
                          XACCINTP1R * LOG( ZVEC2(JJ) ) + XACCINTP2R ) )
#else
                          XACCINTP1R * BR_LOG( ZVEC2(JJ) ) + XACCINTP2R ) )
#endif
      IVEC2(JJ) = INT( ZVEC2(JJ) )
      ZVEC2(JJ) = ZVEC2(JJ) - REAL( IVEC2(JJ) )
    !$mnh_end_do()
    !
    !        5.2.3  perform the bilinear interpolation of the normalized
    !               RACCSS-kernel
    !
!$acc loop independent
    DO JJ = 1, IGACC
      ZVEC3(JJ) =  (  XKER_RACCSS(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RACCSS(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                          * ZVEC1(JJ) &
                 - (  XKER_RACCSS(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    - XKER_RACCSS(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                          * (ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGACC
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    !        5.2.4  raindrop accretion on the small sized aggregates
    !
!$acc loop independent
    DO JL=1, ISIZE
      IF ( GACC(JL) ) THEN
#ifndef MNH_BITREP
        ZZW6(JL) =                                                         & !! coef of RRACCS
              XFRACCSS*( PLBDAS(JL)**XCXS )*( PRHODREF(JL)**(-XCEXVT-1.) ) &
         *( XLBRACCS1/((PLBDAS(JL)**2)                ) +                  &
            XLBRACCS2/( PLBDAS(JL)    * PLBDAR(JL)    ) +                  &
            XLBRACCS3/(               (PLBDAR(JL)**2)) )/PLBDAR(JL)**4
#else
        ZZW6(JL) =                                                         & !! coef of RRACCS
              XFRACCSS* BR_POW(PLBDAS(JL),XCXS) * BR_POW(PRHODREF(JL),-XCEXVT-1.) &
         *( XLBRACCS1/BR_P2(PLBDAS(JL))                 +                  &
            XLBRACCS2/( PLBDAS(JL)    * PLBDAR(JL)    ) +                  &
            XLBRACCS3/               (BR_P2(PLBDAR(JL))) )/BR_P4(PLBDAR(JL))
#endif
        PRS_TEND(JL, IRRACCSS) =ZZW(JL)*ZZW6(JL)
      END IF
    END DO
    !
    !        5.2.4b perform the bilinear interpolation of the normalized
    !               RACCS-kernel
    !
!$acc loop independent
    DO JJ = 1, IGACC
      ZVEC3(JJ) =  (   XKER_RACCS(IVEC1(JJ)+1,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    -  XKER_RACCS(IVEC1(JJ)+1,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                                   * ZVEC1(JJ) &
                 - (   XKER_RACCS(IVEC1(JJ)  ,IVEC2(JJ)+1)* ZVEC2(JJ)          &
                    -  XKER_RACCS(IVEC1(JJ)  ,IVEC2(JJ)  )*(ZVEC2(JJ) - 1.0) ) &
                                                           * (ZVEC1(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGACC
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    WHERE(GACC(:))
      PRS_TEND(:, IRRACCS) = ZZW(:)*ZZW6(:)
    END WHERE
    !        5.2.5  perform the bilinear interpolation of the normalized
    !               SACCRG-kernel
    !
!$acc loop independent
    DO JJ = 1, IGACC
        ZVEC3(JJ) =  (  XKER_SACCRG(IVEC2(JJ)+1,IVEC1(JJ)+1)* ZVEC1(JJ)          &
                      - XKER_SACCRG(IVEC2(JJ)+1,IVEC1(JJ)  )*(ZVEC1(JJ) - 1.0) ) &
                                                            * ZVEC2(JJ) &
                   - (  XKER_SACCRG(IVEC2(JJ)  ,IVEC1(JJ)+1)* ZVEC1(JJ)          &
                      - XKER_SACCRG(IVEC2(JJ)  ,IVEC1(JJ)  )*(ZVEC1(JJ) - 1.0) ) &
                                                            * (ZVEC2(JJ) - 1.0)
    END DO
    ZZW(:) = 0.
!$acc loop independent
    DO JJ = 1, IGACC
      ZZW(I1(JJ)) = ZVEC3(JJ)
    END DO
    !
    !        5.2.6  raindrop accretion-conversion of the large sized aggregates
    !               into graupeln
    !
    WHERE(GACC(:))
#ifndef MNH_BITREP
      PRS_TEND(:, IRSACCRG) = XFSACCRG*ZZW(:)*                    & ! RSACCRG
          ( PLBDAS(:)**(XCXS-XBS) )*( PRHODREF(:)**(-XCEXVT-1.) ) &
         *( XLBSACCR1/((PLBDAR(:)**2)               ) +           &
            XLBSACCR2/( PLBDAR(:)    * PLBDAS(:)    ) +           &
            XLBSACCR3/(               (PLBDAS(:)**2)) )/PLBDAR(:)
#else
      PRS_TEND(:, IRSACCRG) = XFSACCRG*ZZW(:)*                          & ! RSACCRG
            BR_POW(PLBDAS(:),XCXS-XBS) * BR_POW(PRHODREF(:),-XCEXVT-1.) &
         *( XLBSACCR1/BR_P2(PLBDAR(:))                +                 &
            XLBSACCR2/( PLBDAR(:)    * PLBDAS(:)    ) +                 &
            XLBSACCR3/               BR_P2(PLBDAS(:)) )/PLBDAR(:)
#endif
    END WHERE
!$acc end kernels
  ENDIF
ENDIF
!$acc kernels
!
!$acc loop independent
DO JL=1, ISIZE
  ! More restrictive ACC mask to be used for accretion by negative temperature only
  ZACC(JL) = ZACC(JL) * &
           &MAX(0., -SIGN(1., PT(JL)-XTT)) ! WHERE(PT(:)<XTT)
  PRRACCSS(JL)=ZACC(JL)*MIN(ZFREEZ_RATE(JL), PRS_TEND(JL, IRRACCSS))
  ZFREEZ_RATE(JL)=MAX(0., ZFREEZ_RATE(JL)-PRRACCSS(JL))
  ZZW(JL) = MIN(1., ZFREEZ_RATE(JL) / MAX(1.E-20, PRS_TEND(JL, IRRACCS)-PRRACCSS(JL))) ! proportion we are able to freeze
  PRRACCSG(JL)=ZACC(JL)*ZZW(JL) * MAX(0., PRS_TEND(JL, IRRACCS)-PRRACCSS(JL))
  ZFREEZ_RATE(JL) = MAX(0., ZFREEZ_RATE(JL)-PRRACCSG(JL))
  PRSACCRG(JL)=ZACC(JL)*ZZW(JL) * PRS_TEND(JL, IRSACCRG)

  PRSACCRG(JL) = PRSACCRG(JL) * MAX(0., -SIGN(1., -PRRACCSG(JL)))
  PRRACCSG(JL)=MAX(0., PRRACCSG(JL))

  PA_RR(JL) = PA_RR(JL) - PRRACCSS(JL)
  PA_RS(JL) = PA_RS(JL) + PRRACCSS(JL)
  PA_TH(JL) = PA_TH(JL) + PRRACCSS(JL)*(PLSFACT(JL)-PLVFACT(JL))
  PA_RR(JL) = PA_RR(JL) - PRRACCSG(JL)
  PA_RS(JL) = PA_RS(JL) - PRSACCRG(JL)
  PA_RG(JL) = PA_RG(JL) + PRRACCSG(JL)+PRSACCRG(JL)
  PA_TH(JL) = PA_TH(JL) + PRRACCSG(JL)*(PLSFACT(JL)-PLVFACT(JL))
ENDDO
!
!
!*       5.3    Conversion-Melting of the aggregates
!
DO JL=1, ISIZE
  ZMASK(JL)=MAX(0., -SIGN(1., XRTMIN(5)-PRST(JL))) * & ! WHERE(PRST(:)>XRTMIN(5))
           &MAX(0., -SIGN(1., XTT-PT(JL))) * & ! WHERE(PT(:)>XTT)
           &PCOMPUTE(JL)
ENDDO
IF(LDSOFT) THEN
  DO JL=1, ISIZE
    PRSMLTG(JL)=ZMASK(JL)*PRSMLTG(JL)
    PRCMLTSR(JL)=ZMASK(JL)*PRCMLTSR(JL)
  ENDDO
ELSE
  DO JL=1, ISIZE
    PRSMLTG(JL)=ZMASK(JL)*PRVT(JL)*PPRES(JL)/(XEPSILO+PRVT(JL)) ! Vapor pressure
  ENDDO
  IF(LEVLIMIT) THEN
    WHERE(ZMASK(:)==1.)
#ifndef MNH_BITREP
      PRSMLTG(:)=MIN(PRSMLTG(:), EXP(XALPW-XBETAW/PT(:)-XGAMW*ALOG(PT(:)))) ! min(ev, es_w(T))
#else
      PRSMLTG(:)=MIN(PRSMLTG(:), BR_EXP(XALPW-XBETAW/PT(:)-XGAMW*BR_LOG(PT(:)))) ! min(ev, es_w(T))
#endif
    END WHERE
  ENDIF
  DO JL=1, ISIZE
    PRSMLTG(JL)=ZMASK(JL)*( &
                            & PKA(JL)*(XTT-PT(JL)) +                                 &
                            & ( PDV(JL)*(XLVTT + ( XCPV - XCL ) * ( PT(JL) - XTT )) &
                            & *(XESTT-PRSMLTG(JL))/(XRV*PT(JL))             ) &
                          &)
  ENDDO
  PRCMLTSR(:) = 0.
  WHERE(ZMASK(:)==1.)
    !
    ! compute RSMLT
    !
    PRSMLTG(:)  = XFSCVMG*MAX( 0.0,( -PRSMLTG(:) *             &
#ifndef MNH_BITREP
                         ( X0DEPS*       PLBDAS(:)**XEX0DEPS +     &
                           X1DEPS*PCJ(:)*PLBDAS(:)**XEX1DEPS ) -   &
#else
                         ( X0DEPS*       BR_POW(PLBDAS(:),XEX0DEPS) +     &
                           X1DEPS*PCJ(:)*BR_POW(PLBDAS(:),XEX1DEPS) ) -   &
#endif
                                   ( PRS_TEND(:, IRCRIMS) + PRS_TEND(:, IRRACCS) ) *       &
                            ( PRHODREF(:)*XCL*(XTT-PT(:))) ) /    &
                                           ( PRHODREF(:)*XLMTT ) )
    ! When T < XTT, rc is collected by snow (riming) to produce snow and graupel
    ! When T > XTT, if riming was still enabled, rc would produce snow and graupel with snow becomming graupel (conversion/melting) and graupel becomming rain (melting)
    ! To insure consistency when crossing T=XTT, rc collected with T>XTT must be transformed in rain.
    ! rc cannot produce iced species with a positive temperature but is still collected with a good efficiency by snow
    PRCMLTSR(:) = PRS_TEND(:, IRCRIMS) ! both species are liquid, no heat is exchanged
  END WHERE
ENDIF

DO JL=1, ISIZE
  ! note that RSCVMG = RSMLT*XFSCVMG but no heat is exchanged (at the rate RSMLT)
  ! because the graupeln produced by this process are still icy!!!
  PA_RS(JL) = PA_RS(JL) - PRSMLTG(JL)
  PA_RG(JL) = PA_RG(JL) + PRSMLTG(JL)
  PA_RC(JL) = PA_RC(JL) - PRCMLTSR(JL)
  PA_RR(JL) = PA_RR(JL) + PRCMLTSR(JL)
ENDDO
!$acc end kernels

IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK1D(PRSMLTG,"ICE4_FAST_RS end:PRSMLTG",PRECISION)
  CALL MPPDB_CHECK1D(PRCMLTSR,"ICE4_FAST_RS end:PRCMLTSR",PRECISION)
  CALL MPPDB_CHECK2D(PRS_TEND,"ICE4_FAST_RS end:PRS_TEND",PRECISION)
  CALL MPPDB_CHECK1D(PA_TH,"ICE4_FAST_RS end:PA_TH",PRECISION)
  CALL MPPDB_CHECK1D(PA_RC,"ICE4_FAST_RS end:PA_RC",PRECISION)
  CALL MPPDB_CHECK1D(PA_RR,"ICE4_FAST_RS end:PA_RR",PRECISION)
  CALL MPPDB_CHECK1D(PA_RS,"ICE4_FAST_RS end:PA_RS",PRECISION)
  CALL MPPDB_CHECK1D(PA_RG,"ICE4_FAST_RS end:PA_RG",PRECISION)
  !Check all OUT arrays
  CALL MPPDB_CHECK1D(PRCRIMSS,"ICE4_FAST_RS end:PRCRIMSS",PRECISION)
  CALL MPPDB_CHECK1D(PRCRIMSG,"ICE4_FAST_RS end:PRCRIMSG",PRECISION)
  CALL MPPDB_CHECK1D(PRSRIMCG,"ICE4_FAST_RS end:PRSRIMCG",PRECISION)
  CALL MPPDB_CHECK1D(PRRACCSS,"ICE4_FAST_RS end:PRRACCSS",PRECISION)
  CALL MPPDB_CHECK1D(PRRACCSG,"ICE4_FAST_RS end:PRRACCSG",PRECISION)
  CALL MPPDB_CHECK1D(PRSACCRG,"ICE4_FAST_RS end:PRSACCRG",PRECISION)
END IF

!$acc end data

#ifndef MNH_OPENACC
deallocate( zfreez_rate )
deallocate( zzw6        )
deallocate( zzw2        )
deallocate( zzw         )
deallocate( zvec3       )
deallocate( zvec2       )
deallocate( zvec1       )
deallocate( zmask       )
deallocate( zacc        )
deallocate( zrim        )
deallocate( gacc        )
deallocate( grim        )
deallocate( ivec2       )
deallocate( ivec1       )
deallocate( i1          )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

!$acc end data

END SUBROUTINE ICE4_FAST_RS
